@@ -23,7 +23,7 @@ import time
 from calculate.core.server.gen_pid import search_worked_process
 from calculate.core.setup_cache import Cache as SetupCache
 from calculate.core.server.func import MethodsInterface
-from calculate.lib.cl_template import SystemIni
+from calculate.lib.cl_template import SystemIni, LayeredIni
 from calculate.lib.datavars import DataVarsError
 
 from calculate.lib.utils.tools import AddonError
@@ -32,15 +32,15 @@ from calculate.lib.utils.colortext import get_color_print
 from calculate.update.emerge_parser import RevdepPercentBlock
 from calculate.update.datavars import DataVarsUpdate
 from calculate.update.update_info import UpdateInfo
+from calculate.lib.utils.binhosts import fetch_packages
 from calculate.lib.cl_log import log
 import re
 import shutil
 from collections import MutableSet
-from update_tasks import EmergeMark
 
 from calculate.lib.utils.git import Git, GitError, MTimeKeeper, NotGitError
 from calculate.lib.utils.portage import (Layman, EmergeLog,
-                                         EmergeLogNamedTask, PackageList,
+                                         EmergeLogNamedTask,
                                          PackageInformation,
                                          get_packages_files_directory,
                                          get_manifest_files_directory,
@@ -50,7 +50,8 @@ from calculate.lib.utils.text import _u8
 Colors = TextState.Colors
 from calculate.lib.utils.files import (getProgPath, STDOUT, removeDir,
                                        PercentProgress, process, getRunCommands,
-                                       readFile, listDirectory)
+                                       readFile, listDirectory, pathJoin,
+                                       writeFile)
 import emerge_parser
 import logging
 from emerge_parser import (EmergeParser, EmergeCommand, EmergeError,
@@ -76,14 +77,14 @@ class OverlayOwnCache(MutableSet):
         self.dv = dv
 
     def __get_overlays(self):
-        own_cache_value = SystemIni(self.dv).getVar('update', 'own_cache') or ""
+        own_cache_value = SystemIni(self.dv).getVar('system', 'own_cache') or ""
         return [x.strip() for x in own_cache_value.split(',') if x.strip()]
 
     def __write_overlays(self, overlays):
         if not overlays:
-            SystemIni(self.dv).delVar('update', 'own_cache')
+            SystemIni(self.dv).delVar('system', 'own_cache')
         else:
-            SystemIni(self.dv).setVar('update',
+            SystemIni(self.dv).setVar('system',
                                       {'own_cache': ",".join(overlays)})
 
     def __contains__(self, item):
@@ -146,12 +147,14 @@ class Update(MethodsInterface):
                 self.clVars.Get('update.cl_update_rep_path')))
         self.update_map = {}
         self.refresh_binhost = False
+        self.pkgnum = None
+        self.pkgnummax = None
 
     def get_prog_path(self, program_name):
         return getProgPath(program_name)
 
     def _syncRepository(self, name, url, rpath, revision,
-                        cb_progress=None, clean=False):
+                        cb_progress=None, clean=False, notask=False):
         """
         Синхронизировать репозитори
         """
@@ -164,6 +167,10 @@ class Update(MethodsInterface):
         try:
             self.stash_cache(rpath, name)
             if not git.checkExistsRep(rpath):
+                if not notask:
+                    self.startTask(_("Syncing the {rep} repository").format(
+                        rep=name.capitalize()))
+                self.addProgress()
                 git.cloneTagRepository(url, rpath, revision,
                                        cb_progress=cb_progress)
                 info_outdated = True
@@ -174,10 +181,7 @@ class Update(MethodsInterface):
                     tag_cr = git.getCommit(rpath, revision)
                     cr = git.getCurrentCommit(rpath)
                     ref_type = git.reference_type(rpath, revision)
-                    if git.isNeedUnpack(rpath):
-                        need_update = True
-                        clean = True
-                    elif tag_cr != cr or ref_type == Git.Reference.Branch:
+                    if tag_cr != cr or ref_type == Git.Reference.Branch:
                         need_update = True
                     elif clean:
                         status = git.getStatusInfo(rpath)
@@ -186,6 +190,10 @@ class Update(MethodsInterface):
                 except GitError:
                     need_update = True
                 if need_update:
+                    if not notask:
+                        self.startTask(_("Syncing the {rep} repository").format(
+                            rep=name.capitalize()))
+                    self.addProgress()
                     git.updateTagRepository(url, rpath, revision,
                                             cb_progress=cb_progress,
                                             clean=clean)
@@ -222,7 +230,7 @@ class Update(MethodsInterface):
         if not status:
             self.printWARNING(_("Updates autocheck is not enabled"))
             return False
-        last_check = SystemIni().getVar('update', 'last_check') or ""
+        last_check = SystemIni(self.clVars).getVar('system', 'last_check') or ""
         re_interval = re.compile("^(\d+)\s*(hours?|days?|weeks?)?", re.I)
         interval_match = re_interval.search(interval)
         MINUTE = 60
@@ -297,7 +305,8 @@ class Update(MethodsInterface):
         return True
 
     @variable_module("update")
-    def syncRepositories(self, repname, clean_on_error=True):
+    def syncRepositories(self, repname, fallback_sync=False,
+                         clean_on_error=True):
         """
         Синхронизировать репозитории
         """
@@ -318,7 +327,6 @@ class Update(MethodsInterface):
             rpath_orig = rpath
         else:
             rpath_orig = rpath[len(chroot_path):]
-        self.addProgress()
 
         mtime = MTimeKeeper(path.join(rpath, "profiles/updates"))
         mtime.save()
@@ -333,7 +341,8 @@ class Update(MethodsInterface):
                         layman.add(repname, url, rpath_orig)
                     if not self._syncRepository(repname, url, rpath, revision,
                                                 cb_progress=self.setProgress,
-                                                clean=check_status):
+                                                clean=check_status,
+                                                notask=fallback_sync):
                         return "skip"
                     return True
                 except GitError as e:
@@ -350,7 +359,8 @@ class Update(MethodsInterface):
                     try:
                         self._syncRepository(repname, url, rpath_new, revision,
                                              cb_progress=self.setProgress,
-                                             clean=check_status)
+                                             clean=check_status,
+                                             notask=fallback_sync)
                         removeDir(rpath)
                         shutil.move(rpath_new, rpath)
                     except OSError as e:
@@ -559,6 +569,10 @@ class Update(MethodsInterface):
         """
         self.endTask()
         _print = self.color_print
+        if self.pkgnum is not None:
+            num = self.pkgnum
+        if self.pkgnummax is not None:
+            max_num = self.pkgnummax
         one = _print("{0}", num)
         two = _print("{0}", max_num)
         part = _("({current} of {maximum})").format(current=one, maximum=two)
@@ -750,37 +764,12 @@ class Update(MethodsInterface):
                 _("Listing packages for removal")))
             self._display_pretty_package_list(pkglist, remove_list=True)
 
-    def getCacheOnWorld(self, params, packages, check=False):
-        """
-        Получить список обновляемых пакетов @world из кэша
-        """
-        if "@world" in packages:
-            elog = EmergeLog(EmergeLogNamedTask(EmergeMark.Premerge))
-            if check and (elog.list or elog.remove_list):
-                self.emerge_cache.drop_cache(
-                    "Some packages was installed or removed")
-                return params, packages
-            installed_pkgs = elog.list
-            new_packages = self.emerge_cache.get_cached_package_list()
-            if new_packages is not None:
-                return "-1O", ["=%s" % x for x in new_packages
-                               if not str(x) in installed_pkgs]
-        return params, packages
-
-    def updateCache(self, pkg_list):
-        """
-        Обновить кэш. Оставить отметку в emerge.log о том, выполнено действие
-        premerge
-        """
-        self.emerge_cache.set_cache(pkg_list)
-        elog = EmergeLog(EmergeLogNamedTask(EmergeMark.Premerge))
-        elog.mark_end_task(),
 
     def mark_schedule(self):
         """
         Установить отметку о запуске запланированной проверки
         """
-        SystemIni().setVar('update', {'last_check': str(int(time.time()))})
+        SystemIni(self.clVars).setVar('system', {'last_check': str(int(time.time()))})
 
     def get_default_emerge_opts(self):
         return self.clVars.Get('cl_emerge_default_opts')
@@ -790,26 +779,12 @@ class Update(MethodsInterface):
         Вывести информацию об обновлении
         """
         deo = self.clVars.Get('cl_emerge_default_opts')
-        param, packages = self.getCacheOnWorld(param, packages, check=True)
         param = [param, "-pv"]
 
-        if not packages:
-            self.printSUCCESS(_("Installed packages are up to date"))
-            self.set_need_update(False)
-            return True
         with EmergeParser(EmergeCommand(list(packages), emerge_default_opts=deo,
                                         extra_params=param)) as emerge:
             try:
                 emerge.run()
-                if "@world" in packages:
-                    if emerge.install_packages.remove_list:
-                        self.emerge_cache.drop_cache(
-                            "List has packages for remove")
-                    elif emerge.install_packages.block_packages:
-                        self.emerge_cache.drop_cache(
-                            "List has block packages")
-                    else:
-                        self.updateCache(emerge.install_packages.list)
                 if not emerge.install_packages.list:
                     self.printSUCCESS(_("The system is up to date"))
                     self.set_need_update(False)
@@ -820,7 +795,6 @@ class Update(MethodsInterface):
                     self._display_error(emerge.skipped_packages)
             except EmergeError:
                 self.set_need_update(False)
-                self.emerge_cache.drop_cache("Emerge error")
                 self._display_install_package(emerge, emergelike=True)
                 self._display_error(emerge.prepare_error)
                 raise
@@ -855,7 +829,7 @@ class Update(MethodsInterface):
         """
         Установить кэш - "нет пакетов для обновления"
         """
-        self.updateCache(PackageList([]))
+        #self.updateCache(PackageList([]))
         return True
 
     def _startEmerging(self, emerge):
@@ -881,7 +855,6 @@ class Update(MethodsInterface):
         try:
             emerge.run()
         except EmergeError:
-            self.emerge_cache.drop_cache("Emerge error")
             if emerge.emerging_error:
                 self._display_error(emerge.emerging_error.log)
             else:
@@ -902,9 +875,6 @@ class Update(MethodsInterface):
             packages = [param]
             extra_params = None
         else:
-            param, packages = self.getCacheOnWorld(param, packages)
-            if not packages:
-                return True
             extra_params = [param]
 
         command = EmergeCommand(list(packages), emerge_default_opts=deo,
@@ -920,7 +890,43 @@ class Update(MethodsInterface):
                 if not emerge.install_packages.list:
                     return True
             except EmergeError:
-                self.emerge_cache.drop_cache("Emerge error")
+                self._display_error(emerge.prepare_error)
+                raise
+            self._startEmerging(emerge)
+        return True
+
+    def emerge_ask(self, pretend, *params):
+        """
+        Вывести информацию об обновлении
+        """
+        deo = self.get_default_emerge_opts()
+        param = [x for x in params if x.startswith("-")]
+        packages = [x for x in params if not x.startswith("-")]
+        command = EmergeCommand(list(packages), emerge_default_opts=deo,
+                                extra_params=param)
+        with EmergeParser(command) as emerge:
+            try:
+                emerge.question.action = lambda x: False
+                emerge.run()
+                if emerge.install_packages.list:
+                    emergelike = self.clVars.Get(
+                        'update.cl_update_emergelist_set') == 'on'
+                    self._display_install_package(emerge, emergelike)
+                    if emerge.skipped_packages:
+                        self._display_error(emerge.skipped_packages)
+                    if not pretend:
+                        answer = self.askConfirm(
+                            _("Would you like to merge these packages?"), "yes")
+                        if answer == "no":
+                            emerge.command.send("no\n")
+                            raise KeyboardInterrupt
+                    else:
+                        return True
+                else:
+                    self.printSUCCESS(_("The system is up to date"))
+            except EmergeError:
+                self.set_need_update(False)
+                self._display_install_package(emerge, emergelike=True)
                 self._display_error(emerge.prepare_error)
                 raise
             self._startEmerging(emerge)
@@ -1173,20 +1179,34 @@ class Update(MethodsInterface):
         :return:
         """
         hosts = self.clVars.Get("update.cl_update_binhost_host")
+        datas = self.clVars.Get("update.cl_update_binhost_revisions")
+        bin_cache_fn = pathJoin(self.clVars.Get('cl_chroot_path'),
+                                LayeredIni.IniPath.Grp)
         if not hosts:
             self.clVars.Delete('cl_update_binhost', location="system")
+            try:
+                if path.exists(bin_cache_fn):
+                    os.unlink(bin_cache_fn)
+            except OSError:
+                raise UpdateError(
+                    _("Failed to remove cached ini.env of binary repository"))
+
             raise UpdateError(_("Failed to find the server with "
                                 "appropriate updates"))
+        else:
+            with writeFile(bin_cache_fn) as f:
+                f.write(datas[0].strip()+"\n")
         if write_binhost:
             if hosts[0] != self.clVars.Get('update.cl_update_binhost'):
                 self.refresh_binhost = True
+                self.clVars.Set('cl_update_package_cache_set', 'on')
             self.clVars.Write('cl_update_binhost', hosts[0], location="system")
             new_ts = self.clVars.Get("update.cl_update_binhost_timestamp")
             if new_ts:
                 new_ts = new_ts[0]
                 if new_ts.isdigit():
                     ini = SystemIni(self.clVars)
-                    ini.setVar('update', {'last_update': new_ts})
+                    ini.setVar('system', {'last_update': new_ts})
         if self.clVars.Get('cl_action') == 'sync':
             value = self.clVars.GetBool('cl_update_binhost_stable_set')
             new_value = self.clVars.GetBool('cl_update_binhost_stable_opt_set')
@@ -1246,3 +1266,9 @@ class Update(MethodsInterface):
         dv.Set('update.cl_update_branch_name', revs)
         dv.Invalidate('update.cl_update_rep_rev')
         return True
+
+    def download_packages(self, url_binhost, packages_fn):
+        fetch_packages(url_binhost, packages_fn)
+        return True
+
+
